﻿/*
 * Copyright 2021 faddenSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

using System;
using System.Collections.Generic;

using PluginCommon;

namespace Metroid {
    /// <summary>
    /// Helper plugin for Metroid.
    ///
    /// The game does a JSR to ChooseRoutine, which converts the value in the accumulator
    /// to an index into the 16-bit addresses that immediately follow the JSR, and jumps
    /// to that address.  There's no indication of count or list end, so we can't format
    /// the addresses automatically.  We can mark the JSR as "no continue" though, so that
    /// the analyzer doesn't try to treat the addresses as code.
    ///
    /// The "EndPPUString" function at +01c386 also does not return.  It's only called from
    /// a few places, but we might as well mark it while we're here.
    /// </summary>
    public class MetroidHelp : MarshalByRefObject, IPlugin, IPlugin_SymbolList,
            IPlugin_InlineJsr {
        private IApplication mAppRef;
        private byte[] mFileData;
        private AddressTranslate mAddrTrans;

        // One "choose routine" function, but multiple entry points via jump table.  Use as prefix.
        private const string CALL_PREFIX = "ChooseRoutine";

        // One "end PPU string" function.
        private const string ENDPPU_NAME = "EndPPUString";

        private enum InlineKind { Unknown = 0, ChooseRoutine, EndPPU };

        // Map of JSR offsets in project to inline data type expected to follow.
        private Dictionary<int, InlineKind> mInlineOffsets = new Dictionary<int, InlineKind>();

        // List of "interesting" addresses.  Used as an optimization.
        private Dictionary<int, int> mInlineAddrs = new Dictionary<int, int>();

        public string Identifier {
            get {
                return "Metroid Helper";
            }
        }

        public void Prepare(IApplication appRef, byte[] fileData, AddressTranslate addrTrans) {
            mAppRef = appRef;
            mFileData = fileData;
            mAddrTrans = addrTrans;

            mAppRef.DebugLog("MetroidHelp(id=" + AppDomain.CurrentDomain.Id + "): prepare()");
        }

        public void Unprepare() {
            mAppRef = null;
            mFileData = null;
            mAddrTrans = null;
        }

        // IPlugin_SymbolList
        public void UpdateSymbolList(List<PlSymbol> plSyms) {
            mInlineOffsets.Clear();
            mInlineAddrs.Clear();

            foreach (PlSymbol sym in plSyms) {
                if (sym.Value == AddressTranslate.NON_ADDR || sym.Offset < 0) {
                    continue;
                }
                if (sym.Label.StartsWith(CALL_PREFIX)) {
                    // Offsets will be unique.
                    mInlineOffsets.Add(sym.Offset, InlineKind.ChooseRoutine);
                    // Symbol values (addresses) may not be unique.
                    mInlineAddrs[sym.Value] = sym.Value;
                } else if (sym.Label == ENDPPU_NAME) {
                    mInlineOffsets.Add(sym.Offset, InlineKind.EndPPU);
                    mInlineAddrs[sym.Value] = sym.Value;
                }
            }
            mAppRef.DebugLog("Found matches for " + mInlineOffsets.Count + " labels");
        }
        // IPlugin_SymbolList
        public bool IsLabelSignificant(string beforeLabel, string afterLabel) {
            return beforeLabel.StartsWith(CALL_PREFIX) || afterLabel.StartsWith(CALL_PREFIX) ||
                beforeLabel == ENDPPU_NAME || afterLabel == ENDPPU_NAME;
        }

        public void CheckJsr(int offset, int operand, out bool noContinue) {
            noContinue = false;

            // Do a quick test on the target address.
            int unused;
            if (!mInlineAddrs.TryGetValue(operand, out unused)) {
                // JSR destination address not special.
                return;
            }

            // Address matched.  Translate the target address to the actual offset.  This is
            // important when multiple offsets have the same address.
            int targetOffset = mAddrTrans.AddressToOffset(offset, operand);
            if (targetOffset < 0) {
                mAppRef.DebugLog("Failed to map address $" + operand.ToString("x4") + " to offset");
                return;
            }
            InlineKind kind;
            if (!mInlineOffsets.TryGetValue(targetOffset, out kind)) {
                // Actual call target doesn't have a matching label.
                return;
            }

            // Both kinds just set the "no continue" flag.
            noContinue = true;
        }
    }
}
